import asyncio
import logging.config
import os
import resource
import sys
from collections.abc import Callable, Coroutine
from functools import wraps
from json import JSONDecodeError
from typing import Any, TypeVar

import click
import uvicorn
from botocore.exceptions import ClientError, NoCredentialsError
from pydantic import ValidationError

from sagemaker_shim.app import app
from sagemaker_shim.logging import LOGGING_CONFIG
from sagemaker_shim.models import (
    AuxiliaryData,
    InferenceTaskList,
    get_s3_file_content,
    get_s3_resources,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


def async_to_sync(
    func: Callable[..., Coroutine[Any, Any, T]],
) -> Callable[..., T]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


@click.group()
def cli() -> None:
    logging.config.dictConfig(LOGGING_CONFIG)
    set_memory_limits()


@cli.command(short_help="Start the model server")
def serve() -> None:
    logger.info("Starting the model server")

    uvicorn.run(
        app=app,
        host="0.0.0.0",
        port=8080,
        log_config=None,
        workers=1,
        # uvloop does not accept the user or group parameters
        # to subprocess so force using asyncio
        loop="asyncio",
    )

    logger.info("Model server stopped")


@cli.command(short_help="Invoke the model")
@click.option(
    "-t",
    "--tasks",
    default=None,
    help="A JSON string of task definitions",
    type=str,
)
@click.option(
    "-f",
    "--file",
    default=None,
    help="S3 URI of a JSON file containing a list of task definitions",
    type=str,
)
@async_to_sync
async def invoke(tasks: str, file: str) -> None:
    logger.info("Invoking the model")

    tasks_json: str | bytes

    async with get_s3_resources() as s3_resources:
        if tasks and file:
            raise click.UsageError("Only one of tasks or file should be set")
        elif tasks:
            tasks_json = tasks
        elif file:
            try:
                tasks_json = await get_s3_file_content(
                    s3_uri=file, s3_resources=s3_resources
                )
            except (ValueError, ClientError, NoCredentialsError) as error:
                raise click.BadParameter(
                    f"The value provided for file is invalid:\n\n{error}"
                ) from error
        else:
            raise click.UsageError("One of tasks or file should be set")

        try:
            parsed_tasks = InferenceTaskList.model_validate_json(tasks_json)
        except (ValidationError, JSONDecodeError) as error:
            raise click.BadParameter(
                f"The tasks definition is invalid:\n\n{error}"
            ) from error

        if parsed_tasks.root:
            async with AuxiliaryData(s3_resources=s3_resources):
                for task in parsed_tasks.root:
                    # Only run one task at a time
                    await task.invoke(s3_resources=s3_resources)
        else:
            raise click.UsageError("Empty task list provided")

    logger.info("Model invocation complete")


def set_memory_limits() -> None:
    max_memory_mb = int(
        os.environ.get("GRAND_CHALLENGE_COMPONENT_MAX_MEMORY_MB", "0")
    )

    if max_memory_mb:
        logger.info(f"Setting memory limit to {max_memory_mb} MB")
        limit = max_memory_mb * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_DATA, (limit, limit))
    else:
        logger.info("Not setting a memory limit")


if __name__ == "__main__":
    # https://pyinstaller.org/en/stable/runtime-information.html#run-time-information
    we_are_bundled = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")

    if we_are_bundled:
        # https://pyinstaller.org/en/stable/runtime-information.html#using-sys-executable-and-sys-argv-0
        cli(sys.argv[1:])
    else:
        cli()



from uuid import uuid4
from zipfile import ZipFile

import pytest

from sagemaker_shim.exceptions import ZipExtractionError
from sagemaker_shim.extract import safe_extract


def test_extract_with_dodgy_file(tmp_path):
    file = tmp_path / "test.zip"
    dest = tmp_path / "dest"

    with ZipFile(file=file, mode="w") as zip:
        zip.writestr("../foo.txt", "hello!")
        zip.writestr("../../foo.txt", "hello!")

    dest.mkdir()

    with pytest.raises(ZipExtractionError) as error:
        safe_extract(src=file, dest=dest)

    assert str(error) == (
        "<ExceptionInfo "
        "ZipExtractionError('Zip file contains invalid paths') tblen=2>"
    )


def test_directory_must_exist(tmp_path):
    with pytest.raises(RuntimeError) as error:
        safe_extract(src=tmp_path, dest=tmp_path / "dfg")

    assert (
        str(error)
        == "<ExceptionInfo RuntimeError('The destination must exist') tblen=2>"
    )


def test_extraction(tmp_path):
    file = tmp_path / "test.zip"
    dest = tmp_path / "dest"
    dest.mkdir()

    with ZipFile(file=file, mode="w") as zip:
        for arcname in [
            "common/base.txt",
            "common/sub/sub.txt",
            "__MACOSX",
            ".DS_Store",
            "desktop.ini",
        ]:
            zip.writestr(arcname, "hello")

        zip.writestr("common/bigfile.bin", "a" * int(8192 * 1.5))

    safe_extract(src=file, dest=dest)

    assert {*dest.rglob("**/*")} == {
        dest / "base.txt",
        dest / "bigfile.bin",
        dest / "sub",
        dest / "sub/sub.txt",
    }

    with open(dest / "bigfile.bin") as f:
        contents = f.readlines()

    assert contents == ["a" * int(8192 * 1.5)]


def test_single_file(tmp_path):
    file = tmp_path / "test.zip"
    dest = tmp_path / "dest"
    dest.mkdir()
    pk = uuid4()

    with ZipFile(file=file, mode="w") as zip:
        zip.writestr("test.txt", str(pk))

    safe_extract(src=file, dest=dest)

    with open(dest / "test.txt") as f:
        content = f.readlines()

    assert content == [str(pk)]


def test_single_file_nested(tmp_path):
    file = tmp_path / "test.zip"
    dest = tmp_path / "dest"
    dest.mkdir()
    pk = uuid4()

    with ZipFile(file=file, mode="w") as zip:
        zip.writestr("foo/test.txt", str(pk))

    safe_extract(src=file, dest=dest)

    with open(dest / "test.txt") as f:
        content = f.readlines()

    assert content == [str(pk)]


def test_single_directory(tmp_path):
    file = tmp_path / "test.zip"
    dest = tmp_path / "dest"
    dest.mkdir()

    with ZipFile(file=file, mode="w") as zip:
        zip.mkdir("just-a-directory")

    safe_extract(src=file, dest=dest)

    assert {*dest.rglob("**/*")} == set()



import copy
import sys
from contextlib import contextmanager
from importlib.metadata import version
from time import sleep
from uuid import uuid4

import docker
import httpx
import pytest

from tests import __version__
from tests.conftest import minio_container
from tests.utils import (
    encode_b64j,
    get_image_config,
    get_new_env_vars,
    mutate_image,
    pull_image,
    push_image,
)

# Tests for compatability with
# https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-batch-code.html


@pytest.fixture(autouse=True)
def _container_helper(request) -> None:
    marker = request.node.get_closest_marker("container")
    if marker:
        request.getfixturevalue("container")  # pragma: no cover


@contextmanager
def _container(*, base_image="ubuntu:latest", host_port=8080, cmd=None):
    client = docker.from_env()
    registry = client.containers.run(
        image="registry:2",
        ports={5000: None},
        auto_remove=True,
        detach=True,
        init=True,
    )
    # Wait for startup
    sleep(1)

    try:
        registry.reload()  # required to get ports
        port = registry.ports["5000/tcp"][0]["HostPort"]
        repo = f"localhost:{port}"
        repo_tag = f"{repo}/{base_image}"

        # Pull the base image
        container_image = client.images.pull(base_image)

        # Tag and push it to the local registry
        container_image.tag(repo_tag)
        push_image(client=client, repo_tag=repo_tag)

        config = get_image_config(repo_tag=repo_tag)
        env_vars = get_new_env_vars(existing_config=config)

        if cmd is not None:
            env_vars["GRAND_CHALLENGE_COMPONENT_CMD_B64J"] = encode_b64j(
                val=cmd
            )

        new_tag = mutate_image(
            repo_tag=repo_tag,
            env_vars=env_vars,
            version=__version__,
        )
        pull_image(client=client, repo_tag=new_tag)

        with minio_container() as minio:
            container_env = copy.deepcopy(minio.env)

            container_env["AWS_S3_ENDPOINT_URL"] = "http://minio:9000"

            container = client.containers.run(
                image=new_tag,
                # For batch transforms, SageMaker runs the container as
                # docker run image serve
                command="serve",
                # Containers must implement a web server that responds
                # to invocations and ping requests on port 8080.
                ports={8080: host_port},
                auto_remove=True,
                detach=True,
                # You can't use the init initializer as your entry point
                # in SageMaker containers because it gets confused by the
                # train and serve arguments
                init=False,
                environment=container_env,
                links={minio.container.name: "minio"},
                user=0,
            )

            # Wait for startup
            sleep(5)

            try:
                yield container
            finally:
                container.stop(timeout=0)
    finally:
        registry.stop(timeout=0)


@pytest.fixture(scope="session")
def container():
    with _container() as c:
        yield c


@pytest.mark.container
@pytest.mark.skipif(
    sys.platform != "linux", reason="does not run outside linux"
)
def test_container_responds_to_ping():
    response = httpx.get("http://localhost:8080/ping", timeout=30)

    # SageMaker waits for an HTTP 200 status code and an empty body
    # for a successful ping request before sending an invocations request.
    assert response.status_code == 200
    assert response.content == b""


@pytest.mark.container
@pytest.mark.skipif(
    sys.platform != "linux", reason="does not run outside linux"
)
def test_container_responds_to_execution_parameters():
    response = httpx.get(
        "http://localhost:8080/execution-parameters", timeout=30
    )

    assert response.json() == {
        "MaxConcurrentTransforms": 1,
        "BatchStrategy": "SINGLE_RECORD",
        "MaxPayloadInMB": 6,
    }


@pytest.mark.container
@pytest.mark.skipif(
    sys.platform != "linux", reason="does not run outside linux"
)
def test_invocations_endpoint(minio):
    # To receive inference requests, the container must have a web server
    # listening on port 8080 and must accept POST requests to the
    # /invocations endpoint.
    pk = str(uuid4())
    data = {
        "pk": str(uuid4()),
        "inputs": [],
        "output_bucket_name": minio.output_bucket_name,
        "output_prefix": f"test/{pk}",
    }
    response = httpx.post(
        "http://localhost:8080/invocations", json=data, timeout=30
    )

    # To obtain inferences, Amazon SageMaker sends a POST request to the
    # inference container. The POST request body contains data from
    # Amazon S3. Amazon SageMaker passes the request to the container,
    # and returns the inference result from the container, saving the
    # data from the response to Amazon S3.
    response = response.json()

    assert response["return_code"] == 0
    assert response["outputs"] == []
    assert response["sagemaker_shim_version"] == version("sagemaker-shim")


@pytest.mark.container
@pytest.mark.skipif(
    sys.platform != "linux", reason="does not run outside linux"
)
def test_alpine_image(minio):
    # https://github.com/JonathonReinhart/staticx/issues/143
    host_port = 8081
    with _container(
        base_image="python:3.12-alpine",
        host_port=host_port,
        cmd=["python", "-c", "print('hello_world')"],
    ):
        pk = str(uuid4())
        data = {
            "pk": pk,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"test/{pk}",
        }
        response = httpx.post(
            f"http://localhost:{host_port}/invocations", json=data, timeout=30
        )

        response = response.json()

        assert response["return_code"] == 0
        assert response["outputs"] == []
        assert response["sagemaker_shim_version"] == version("sagemaker-shim")



import hashlib
import hmac
import io
import json
import logging.config
import os
import secrets
from pathlib import Path
from uuid import uuid4
from zipfile import ZipFile

import pytest

from sagemaker_shim.logging import LOGGING_CONFIG
from sagemaker_shim.models import (
    InferenceIO,
    InferenceResult,
    InferenceTask,
    clean_path,
    get_s3_resources,
)
from tests.utils import encode_b64j


@pytest.mark.asyncio
async def test_input_download(minio, tmp_path, monkeypatch):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH",
        str(tmp_path),
    )

    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    task = InferenceTask(
        pk=pk,
        inputs=[
            InferenceIO(
                bucket_name=minio.input_bucket_name,
                bucket_key=f"{prefix}/root.bin",
                relative_path=Path("root.bin"),
            ),
            InferenceIO(
                bucket_name=minio.input_bucket_name,
                bucket_key=f"{prefix}/sub/dir.bin",
                relative_path=Path("sub/dir.bin"),
            ),
        ],
        output_bucket_name=minio.output_bucket_name,
        output_prefix=str(prefix),
    )

    # Prep input bucket
    root_data = os.urandom(8)
    root_f = io.BytesIO(root_data)

    sub_data = os.urandom(8)
    sub_f = io.BytesIO(sub_data)

    async with get_s3_resources() as s3_resources:
        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                root_f, minio.input_bucket_name, f"{prefix}/root.bin"
            )
            await s3_resources.client.upload_fileobj(
                sub_f, minio.input_bucket_name, f"{prefix}/sub/dir.bin"
            )
            response = await s3_resources.client.list_objects_v2(
                Bucket=minio.input_bucket_name,
                Prefix=prefix,
            )

        await task.download_input(s3_resources=s3_resources)

    assert {f["Key"] for f in response["Contents"]} == {
        f"tasks/{task.pk}/root.bin",
        f"tasks/{task.pk}/sub/dir.bin",
    }

    # Check
    with open(tmp_path / "root.bin", "rb") as f:
        created_root = f.read()

    with open(tmp_path / "sub" / "dir.bin", "rb") as f:
        created_sub = f.read()

    assert created_root == root_data
    assert created_sub == sub_data


@pytest.mark.asyncio
async def test_input_decompress(minio, tmp_path, monkeypatch):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH",
        str(tmp_path),
    )
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    task = InferenceTask(
        pk=pk,
        inputs=[
            InferenceIO(
                bucket_name=minio.input_bucket_name,
                bucket_key=f"{prefix}/sub/predictions.zip",
                relative_path=Path("sub/predictions.zip"),
                decompress=True,
            ),
        ],
        output_bucket_name=minio.output_bucket_name,
        output_prefix=str(prefix),
    )

    # Prep input bucket
    sub_f = io.BytesIO()
    with ZipFile(file=sub_f, mode="w") as zip:
        zip.writestr("sdsdaf/test.txt", str(pk))
    sub_f.seek(0)

    async with get_s3_resources() as s3_resources:
        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                sub_f, minio.input_bucket_name, f"{prefix}/sub/predictions.zip"
            )
            response = await s3_resources.client.list_objects_v2(
                Bucket=minio.input_bucket_name,
                Prefix=prefix,
            )

        await task.download_input(s3_resources=s3_resources)

    assert {f["Key"] for f in response["Contents"]} == {
        f"tasks/{task.pk}/sub/predictions.zip",
    }

    # Check
    with open(tmp_path / "sub" / "test.txt") as f:
        created_sub = f.read()

    assert created_sub == str(pk)


@pytest.mark.asyncio
async def test_invoke_with_dodgy_file(
    client, minio, tmp_path, monkeypatch, capsys
):
    input_path = tmp_path / "input"
    linked_input_parent = tmp_path / "linked-input"

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH", str(input_path.absolute())
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT",
        str(linked_input_parent),
    )

    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    data = {
        "pk": pk,
        "inputs": [
            {
                "bucket_name": minio.input_bucket_name,
                "bucket_key": f"{prefix}/sub/dodgy.zip",
                "relative_path": "sub/dodgy.zip",
                "decompress": True,
            }
        ],
        "output_bucket_name": minio.output_bucket_name,
        "output_prefix": prefix,
    }

    input_path = tmp_path / "input"
    output_path = tmp_path / "output"

    input_path.mkdir()
    output_path.mkdir()

    # Prep input bucket
    sub_f = io.BytesIO()
    with ZipFile(file=sub_f, mode="w") as zip:
        zip.writestr("../foo.txt", "hello!")
        zip.writestr("../../foo.txt", "hello!")
    sub_f.seek(0)

    async with get_s3_resources() as s3_resources:
        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                sub_f, minio.input_bucket_name, f"{prefix}/sub/dodgy.zip"
            )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH",
        str(input_path),
    )

    logging.config.dictConfig(LOGGING_CONFIG)

    response = client.post("/invocations", json=data)

    response = response.json()
    assert response["return_code"] == 1

    captured = capsys.readouterr()
    assert captured.err == (
        '{"log": "Zip file contains invalid paths", "level": "ERROR", '
        f'"source": "stderr", "internal": false, "task": "{pk}"}}\n'
    )


@pytest.mark.asyncio
async def test_invoke_with_non_zip(
    client, minio, tmp_path, monkeypatch, capsys
):
    input_path = tmp_path / "input"
    linked_input_parent = tmp_path / "linked-input"

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH", str(input_path.absolute())
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT",
        str(linked_input_parent),
    )

    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    data = {
        "pk": pk,
        "inputs": [
            {
                "bucket_name": minio.input_bucket_name,
                "bucket_key": f"{prefix}/sub/dodgy.zip",
                "relative_path": "sub/dodgy.zip",
                "decompress": True,
            }
        ],
        "output_bucket_name": minio.output_bucket_name,
        "output_prefix": prefix,
    }

    input_path = tmp_path / "input"
    output_path = tmp_path / "output"

    input_path.mkdir()
    output_path.mkdir()

    # Prep input bucket
    sub_data = os.urandom(8)
    sub_f = io.BytesIO(sub_data)
    sub_f.seek(0)

    async with get_s3_resources() as s3_resources:
        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                sub_f, minio.input_bucket_name, f"{prefix}/sub/dodgy.zip"
            )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH",
        str(input_path),
    )

    logging.config.dictConfig(LOGGING_CONFIG)

    response = client.post("/invocations", json=data)

    response = response.json()
    assert response["return_code"] == 1

    captured = capsys.readouterr()
    assert captured.err == (
        '{"log": "Input zip file could not be extracted", "level": "ERROR", '
        f'"source": "stderr", "internal": false, "task": "{pk}"}}\n'
    )


@pytest.mark.asyncio
async def test_output_upload(minio, tmp_path, monkeypatch):
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=minio.output_bucket_name,
        output_prefix=str(prefix),
    )

    # Prep output bucket
    root_data = os.urandom(8)
    with open(tmp_path / "root.bin", "wb") as f:
        f.write(root_data)

    sub_data = os.urandom(8)
    (tmp_path / "sub").mkdir()
    with open(tmp_path / "sub" / "dir.bin", "wb") as f:
        f.write(sub_data)

    # Upload the data
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_OUTPUT_PATH",
        str(tmp_path),
    )

    root_f = io.BytesIO()
    sub_f = io.BytesIO()

    async with get_s3_resources() as s3_resources:
        await task.upload_output(s3_resources=s3_resources)

        async with s3_resources.semaphore:
            response = await s3_resources.client.list_objects_v2(
                Bucket=minio.output_bucket_name,
                Prefix=prefix,
            )
            await s3_resources.client.download_fileobj(
                Fileobj=root_f,
                Bucket=minio.output_bucket_name,
                Key=f"tasks/{task.pk}/root.bin",
            )
            await s3_resources.client.download_fileobj(
                Fileobj=sub_f,
                Bucket=minio.output_bucket_name,
                Key=f"tasks/{task.pk}/sub/dir.bin",
            )

    root_f.seek(0)
    sub_f.seek(0)

    assert {f["Key"] for f in response["Contents"]} == {
        f"tasks/{task.pk}/root.bin",
        f"tasks/{task.pk}/sub/dir.bin",
    }
    assert root_f.read() == root_data
    assert sub_f.read() == sub_data


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,expected_return_code",
    (
        (["echo", "hello"], 0),
        (["bash", "-c", "exit 1"], 1),
    ),
)
async def test_inference_result_upload(
    minio, tmp_path, monkeypatch, cmd, expected_return_code
):
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=minio.output_bucket_name,
        output_prefix=str(prefix),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    serialised_invocation = io.BytesIO()

    async with get_s3_resources() as s3_resources:
        direct_invocation = await task.invoke(s3_resources=s3_resources)

        async with s3_resources.semaphore:
            await s3_resources.client.download_fileobj(
                Fileobj=serialised_invocation,
                Bucket=minio.output_bucket_name,
                Key=f"tasks/{pk}/.sagemaker_shim/inference_result.json",
            )

    assert direct_invocation.return_code == expected_return_code
    assert direct_invocation.pk == pk

    serialised_invocation.seek(0)

    assert direct_invocation == InferenceResult(
        **json.loads(serialised_invocation.read().decode("utf-8"))
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,expected_return_code",
    (
        (["echo", "hello"], 0),
        (["bash", "-c", "exit 1"], 1),
    ),
)
async def test_inference_result_signed(
    minio, tmp_path, monkeypatch, cmd, expected_return_code
):
    pk = str(uuid4())
    signing_key = secrets.token_hex()
    prefix = f"tasks/{pk}"
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=minio.output_bucket_name,
        output_prefix=str(prefix),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_SIGNING_KEY_HEX", signing_key
    )

    async with get_s3_resources() as s3_resources:
        direct_invocation = await task.invoke(s3_resources=s3_resources)

        async with s3_resources.semaphore:
            response = await s3_resources.client.get_object(
                Bucket=minio.output_bucket_name,
                Key=f"tasks/{pk}/.sagemaker_shim/inference_result.json",
            )

    assert direct_invocation.return_code == expected_return_code
    assert direct_invocation.pk == pk

    data = await response["Body"].read()
    meta_sig = response["Metadata"]["signature_hmac_sha256"]

    assert direct_invocation == InferenceResult(
        **json.loads(data.decode("utf-8"))
    )

    calc = hmac.new(
        key=bytes.fromhex(signing_key), msg=data, digestmod=hashlib.sha256
    ).hexdigest()
    assert secrets.compare_digest(calc, meta_sig)

    calc = hmac.new(
        key=b"different-key-used", msg=data, digestmod=hashlib.sha256
    ).hexdigest()
    assert not secrets.compare_digest(calc, meta_sig)


def test_folder_cleanup(tmp_path):
    for dir in ["test", "nested/test"]:
        (tmp_path / dir).mkdir(parents=True)

    for f in ["test", ".test", "test/test", "test/.test", "nested/test/.test"]:
        (tmp_path / f).touch()

    clean_path(path=tmp_path)

    assert [*tmp_path.rglob("**/*")] == []



import io
import json
import resource
from unittest.mock import patch
from uuid import uuid4

import pytest
from click.testing import CliRunner

from sagemaker_shim.cli import async_to_sync, cli
from sagemaker_shim.models import (
    S3Resources,
    get_s3_file_content,
    get_s3_resources,
)
from tests.utils import encode_b64j


def test_invoke_neither_set():
    runner = CliRunner()
    result = runner.invoke(cli, ["invoke"])
    assert result.exit_code == 2
    assert "Error: One of tasks or file should be set" in result.output


def test_invoke_both_set():
    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-f", "fasd", "-t", "[]"])
    assert result.exit_code == 2
    assert "Only one of tasks or file should be set" in result.output


def test_invoke_empty_task_list():
    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-t", "[]"])
    assert result.exit_code == 2
    assert "Empty task list provided" in result.output


@pytest.mark.parametrize(
    "task_list",
    (
        "gdfsa",
        "[{}]",
        "1",
        1,
        "{}",
    ),
)
def test_invoke_invalid_task_list(task_list):
    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-t", task_list])
    assert result.exit_code == 2
    assert "The tasks definition is invalid" in result.output


def test_invoke_missing_s3_file(minio):
    runner = CliRunner()
    result = runner.invoke(
        cli, ["invoke", "-f", f"s3://{minio.output_bucket_name}/missing.json"]
    )
    assert result.exit_code == 2
    assert (
        "An error occurred (404) when calling the HeadObject operation: Not Found"
        in result.output
    )


def test_invoke_bad_bucket(minio):
    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-f", "s3://fasd/missing.json"])
    assert result.exit_code == 2
    assert (
        "An error occurred (404) when calling the HeadObject operation: Not Found"
        in result.output
    )


async def s3_upload_fileobj(*args, s3_resources: S3Resources, **kwargs):
    async with s3_resources.semaphore:
        await s3_resources.client.upload_fileobj(*args, **kwargs)


def sync_s3_operation(*, method, **kwargs):
    # click applications are synchronous and must be tested
    # synchronously. This converts async s3 operations to sync
    # for use in tests.
    @async_to_sync
    async def _run():
        async with get_s3_resources() as s3_resources:
            return await method(**kwargs, s3_resources=s3_resources)

    return _run()


@pytest.mark.parametrize(
    "cmd,expected_return_code",
    (
        (["echo", "hello"], 0),
        (["bash", "-c", "exit 1"], 1),
    ),
)
def test_inference_from_task_list(
    minio, monkeypatch, cmd, expected_return_code
):
    pk1, pk2 = str(uuid4()), str(uuid4())
    tasks = [
        {
            "pk": pk1,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk1}",
        },
        {
            "pk": pk2,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk2}",
        },
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    runner = CliRunner()
    runner.invoke(cli, ["invoke", "-t", json.dumps(tasks)])

    for pk in pk1, pk2:
        serialised_invocation = sync_s3_operation(
            method=get_s3_file_content,
            s3_uri=f"s3://{minio.output_bucket_name}/tasks/{pk}/.sagemaker_shim/inference_result.json",
        )
        parsed_result = json.loads(serialised_invocation)

        assert parsed_result["return_code"] == expected_return_code
        assert parsed_result["pk"] == pk


@pytest.mark.parametrize(
    "cmd,expected_return_code",
    (
        (["echo", "hello"], 0),
        (["bash", "-c", "exit 1"], 1),
    ),
)
def test_inference_from_s3_uri(minio, monkeypatch, cmd, expected_return_code):
    pk1, pk2 = str(uuid4()), str(uuid4())
    tasks = [
        {
            "pk": pk1,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk1}",
        },
        {
            "pk": pk2,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk2}",
        },
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    definition_key = f"{uuid4()}/invocations.json"

    sync_s3_operation(
        method=s3_upload_fileobj,
        Fileobj=io.BytesIO(json.dumps(tasks).encode("utf-8")),
        Bucket=minio.input_bucket_name,
        Key=definition_key,
    )

    runner = CliRunner()
    runner.invoke(
        cli,
        ["invoke", "-f", f"s3://{minio.input_bucket_name}/{definition_key}"],
    )

    for pk in pk1, pk2:
        serialised_invocation = sync_s3_operation(
            method=get_s3_file_content,
            s3_uri=f"s3://{minio.output_bucket_name}/tasks/{pk}/.sagemaker_shim/inference_result.json",
        )
        parsed_result = json.loads(serialised_invocation)

        assert parsed_result["return_code"] == expected_return_code
        assert parsed_result["pk"] == pk


def test_logging_setup(minio, monkeypatch):
    pk = str(uuid4())
    tasks = [
        {
            "pk": pk,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk}",
        }
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-t", json.dumps(tasks)])

    assert (
        '{"log": "hello", "level": "INFO", '
        f'"source": "stdout", "internal": false, "task": "{pk}"}}'
    ) in result.output
    assert "Setting up Auxiliary Data" in result.output
    assert "Cleaning up Auxiliary Data" in result.output


def test_logging_stderr_setup(minio, monkeypatch):
    pk = str(uuid4())
    tasks = [
        {
            "pk": pk,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk}",
        }
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(
            val=["bash", "-c", "echo 'hello' >> /dev/stderr && exit 1"]
        ),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-t", json.dumps(tasks)])

    assert (
        '{"log": "hello", "level": "WARNING", '
        f'"source": "stderr", "internal": false, "task": "{pk}"}}'
    ) in result.output


def test_memory_limit_undefined(minio, monkeypatch):
    pk = str(uuid4())
    tasks = [
        {
            "pk": pk,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk}",
        }
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")

    runner = CliRunner()
    result = runner.invoke(cli, ["invoke", "-t", json.dumps(tasks)])

    assert (
        '{"log": "Not setting a memory limit", "level": "INFO", '
        '"source": "stdout", "internal": true, "task": null}'
    ) in result.output


def test_memory_limit_defined(minio, monkeypatch):
    pk = str(uuid4())
    tasks = [
        {
            "pk": pk,
            "inputs": [],
            "output_bucket_name": minio.output_bucket_name,
            "output_prefix": f"tasks/{pk}",
        }
    ]

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_MAX_MEMORY_MB", "1337")

    expected_limit = 1337 * 1024 * 1024

    with patch("resource.setrlimit") as mock_setrlimit:
        runner = CliRunner()
        result = runner.invoke(cli, ["invoke", "-t", json.dumps(tasks)])

        mock_setrlimit.assert_called_once_with(
            resource.RLIMIT_DATA, (expected_limit, expected_limit)
        )

    assert (
        '{"log": "Setting memory limit to 1337 MB", "level": "INFO", '
        '"source": "stdout", "internal": true, "task": null}'
    ) in result.output
